# the async version is adapted from https://gist.github.com/neubig/80de662fb3e225c18172ec218be4917a

from __future__ import annotations

import os
import yaml
import openai
import ast
import pdb
import asyncio
from typing import Any, List
import os
import pathlib
import openai
import re


# from dotagent.fact_check..env_config import factool_env_config

# env
# openai.api_key = factool_env_config.openai_api_key

class OpenAIChat():
    def __init__(
            self,
            model_name='gpt-3.5-turbo',
            max_tokens=2500,
            temperature=0,
            top_p=1,
            request_timeout=120,
    ):
        if 'gpt' not in model_name:
            openai.api_base = "http://localhost:8000/v1"
        else:
            #openai.api_base = "https://api.openai.com/v1"
            openai.api_key = os.environ.get("OPENAI_API_KEY", None)
            assert openai.api_key is not None, "Please set the OPENAI_API_KEY environment variable."
            assert openai.api_key !='', "Please set the OPENAI_API_KEY environment variable."

        self.config = {
            'model_name': model_name,
            'max_tokens': max_tokens,
            'temperature': temperature,
            'top_p': top_p,
            'request_timeout': request_timeout,
        }

    def extract_list_from_string(self, input_string):
        # pattern = r'\[.*\]'  
        # result = re.search(pattern, input_string)
        # if result:
        #     return result.group()
        # else:
        #     return None
        start_index = input_string.find('[')  
        end_index = input_string.rfind(']') 

        if start_index != -1 and end_index != -1 and start_index < end_index:
            return input_string[start_index:end_index + 1]
        else:
            return None
        
    def extract_dict_from_string(self, input_string):
        start_index = input_string.find('{')
        end_index = input_string.rfind('}')

        if start_index != -1 and end_index != -1 and start_index < end_index:
            return input_string[start_index:end_index + 1]
        else:
            return None
    
    def _boolean_fix(self, output):
        return output.replace("true", "True").replace("false", "False")

    def _type_check(self, output, expected_type):
        try:
            output_eval = ast.literal_eval(output)
            if not isinstance(output_eval, expected_type):
                return None
            return output_eval
        except:
            if(expected_type == List):
                valid_output = self.extract_list_from_string(output)
                output_eval = ast.literal_eval(valid_output)
                if not isinstance(output_eval, expected_type):
                    return None
                return output_eval
            elif(expected_type == dict):
                valid_output = self.extract_dict_from_string(output)
                output_eval = ast.literal_eval(valid_output)
                if not isinstance(output_eval, expected_type):
                    return None
                return output_eval


    async def dispatch_openai_requests(
        self,
        messages_list,
    ) -> list[str]:
        """Dispatches requests to OpenAI API asynchronously.
        
        Args:
            messages_list: List of messages to be sent to OpenAI ChatCompletion API.
        Returns:
            List of responses from OpenAI API.
        """
        async def _request_with_retry(messages, retry=3):
            for _ in range(retry):
                try:
                    response = await openai.ChatCompletion.acreate(
                        model=self.config['model_name'],
                        messages=messages,
                        max_tokens=self.config['max_tokens'],
                        temperature=self.config['temperature'],
                        top_p=self.config['top_p'],
                        request_timeout=self.config['request_timeout'],
                    )
                    return response
                except openai.error.RateLimitError:
                    print('Rate limit error, waiting for 40 second...')
                    await asyncio.sleep(40)
                except openai.error.APIError:
                    print('API error, waiting for 1 second...')
                    await asyncio.sleep(1)
                except openai.error.Timeout:
                    print('Timeout error, waiting for 1 second...')
                    await asyncio.sleep(1)
                except openai.error.ServiceUnavailableError:
                    print('Service unavailable error, waiting for 3 second...')
                    await asyncio.sleep(3)
                except openai.error.APIConnectionError:
                    print('API Connection error, waiting for 3 second...')
                    await asyncio.sleep(3)

            return None

        async_responses = [
            _request_with_retry(messages)
            for messages in messages_list
        ]

        return await asyncio.gather(*async_responses)
    
    async def async_run(self, messages_list, expected_type):
        retry = 1
        responses = [None for _ in range(len(messages_list))]
        messages_list_cur_index = [i for i in range(len(messages_list))]

        while retry > 0 and len(messages_list_cur_index) > 0:
            print(f'{retry} retry left...')
            messages_list_cur = [messages_list[i] for i in messages_list_cur_index]
            
            predictions = await self.dispatch_openai_requests(
                messages_list=messages_list_cur,
            )

            preds = [self._type_check(self._boolean_fix(prediction['choices'][0]['message']['content']), expected_type) if prediction is not None else None for prediction in predictions]

            finised_index = []
            for i, pred in enumerate(preds):
                if pred is not None:
                    responses[messages_list_cur_index[i]] = pred
                    finised_index.append(messages_list_cur_index[i])
            
            messages_list_cur_index = [i for i in messages_list_cur_index if i not in finised_index]
            
            retry -= 1
        
        return responses

class OpenAIEmbed():
    def __init__():
        openai.api_key = os.environ.get("OPENAI_API_KEY", None)
        assert openai.api_key is not None, "Please set the OPENAI_API_KEY environment variable."
        assert openai.api_key != '', "Please set the OPENAI_API_KEY environment variable."

    async def create_embedding(self, text, retry=3):
        for _ in range(retry):
            try:
                response = await openai.Embedding.acreate(input=text, model="text-embedding-ada-002")
                return response
            except openai.error.RateLimitError:
                print('Rate limit error, waiting for 1 second...')
                await asyncio.sleep(1)
            except openai.error.APIError:
                print('API error, waiting for 1 second...')
                await asyncio.sleep(1)
            except openai.error.Timeout:
                print('Timeout error, waiting for 1 second...')
                await asyncio.sleep(1)
        return None

    async def process_batch(self, batch, retry=3):
        tasks = [self.create_embedding(text, retry=retry) for text in batch]
        return await asyncio.gather(*tasks)

if __name__ == "__main__":
    chat = OpenAIChat(model_name='llama-2-7b-chat-hf')

    predictions = asyncio.run(chat.async_run(
        messages_list=[
            [{"role": "user", "content": "show either 'ab' or '['a']'. Do not do anything else."}],
        ] * 20,
        expected_type=List,
    ))

    print(predictions)
    # Usage
    # embed = OpenAIEmbed()
    # batch = ["string1", "string2", "string3", "string4", "string5", "string6", "string7", "string8", "string9", "string10"]  # Your batch of strings
    # embeddings = asyncio.run(embed.process_batch(batch, retry=3))
    # for embedding in embeddings:
    #     print(embedding["data"][0]["embedding"])


"""Definition of different types of tasks."""
from __future__ import annotations

from enum import Enum


class TaskType(str, Enum):
    """Task types available in this tool."""

    kbqa = "kbqa"
    math = "math"
    code = "code"
    sci = "sci"

    @staticmethod
    def list() -> list[str]:
        """Obtains string representations of all values.
        Returns:
            List of all values in str.
        """
        return list(map(lambda c: c.value, TaskType))


__version__ = "0.0.64"

import types
import sys
import os
import requests
from . import library as commands
from ._program import Program
from dotagent import llms

from ._utils import load, merge_programs
from . import selectors
import nest_asyncio
import asyncio

# the user needs to set an LLM before they can use Compiler
llm = None

# This makes the Compiler module callable
class Compiler(types.ModuleType):
    def __call__(self, template, llm=None, cache_seed=0, logprobs=None, silent=None, async_mode=False, stream=None, caching=None, await_missing=False, logging=False, memory=None, memory_threshold=1, **kwargs):
        return Program(template, llm=llm, cache_seed=cache_seed, logprobs=logprobs, silent=silent, async_mode=async_mode, stream=stream, caching=caching, await_missing=await_missing, logging=logging, memory=memory, memory_threshold=memory_threshold, **kwargs)
sys.modules[__name__].__class__ = Compiler


def load(Compiler_file):
    ''' Load a Compiler program from the given text file.

    If the passed file is a valid local file it will be loaded directly.
    Otherwise, if it starts with "http://" or "https://" it will be loaded
    from the web.
    '''

    if os.path.exists(Compiler_file):
        with open(Compiler_file, 'r') as f:
            template = f.read()
    elif Compiler_file.startswith('http://') or Compiler_file.startswith('https://'):
        template = requests.get(Compiler_file).text
    else:
        raise ValueError('Invalid Compiler file: %s' % Compiler_file)
    
    return sys.modules[__name__](template)




import importlib
from typing import Any


def import_class(class_path):
    module_name, class_name = class_path.rsplit(".", 1)
    module = importlib.import_module(module_name)
    return getattr(module, class_name)

def document_loader(reader_type: str) -> Any:
    mapping = {
                "airtable": "dotagent.knowledgebase.document_loaders.airtable.base.AirtableReader",
                "apify_dataset": "dotagent.knowledgebase.document_loaders.apify.dataset.base.ApifyDataset",
                "asana": "dotagent.knowledgebase.document_loaders.asana.base.AsanaReader",
                "azcognitive_search": "dotagent.knowledgebase.document_loaders.azcognitive_search.base.AzCognitiveSearchReader",
                "bilibili": "dotagent.knowledgebase.document_loaders.bilibili.base.BilibiliTranscriptReader",
                "boarddocs": "dotagent.knowledgebase.document_loaders.boarddocs.base.BoardDocsReader",
                "chatgpt_plugin": "dotagent.knowledgebase.document_loaders.chatgpt_plugin.base.ChatGPTRetrievalPluginReader",
                "chroma": "dotagent.knowledgebase.document_loaders.chroma.base.ChromaReader",
                "confluence": "dotagent.knowledgebase.document_loaders.confluence.base.ConfluenceReader",
                "couchdb": "dotagent.knowledgebase.document_loaders.couchdb.base.SimpleCouchDBReader",
                "dad_jokes": "dotagent.knowledgebase.document_loaders.dad_jokes.base.DadJokesReader",
                "deep_lake": "dotagent.knowledgebase.document_loaders.deeplake.base.DeepLakeReader",
                "discord": "dotagent.knowledgebase.document_loaders.discord.base.DiscordReader",
                "docugami": "dotagent.knowledgebase.document_loaders.docugami.base.DocugamiReader",
                "elasticsearch": "dotagent.knowledgebase.document_loaders.elasticsearch.base.ElasticsearchReader",
                "faiss": "dotagent.knowledgebase.document_loaders.faiss.base.FaissReader",
                "feedly_rss": "dotagent.knowledgebase.document_loaders.feedly_rss.base.FeedlyRssReader",
                "feishu_docs": "dotagent.knowledgebase.document_loaders.feishu_docs.base.FeishuDocsReader",
                "file_directory": "dotagent.knowledgebase.document_loaders.file.base.SimpleDirectoryReader",
                "file_audio": "dotagent.knowledgebase.document_loaders.file.audio.base.AudioTranscriber",
                "gladia_audio": "dotagent.knowledgebase.document_loaders.file.audio_gladia.base.GladiaAudioTranscriber",
                "file_cjk_pdf": "dotagent.knowledgebase.document_loaders.file.cjk_pdf.base.CJKPDFReader",
                "deep_doctection": "dotagent.knowledgebase.document_loaders.file.deepdoctection.base.DeepDoctectionReader",
                "file_docx": "dotagent.knowledgebase.document_loaders.file.docx.base.DocxReader",
                "file_epub": "dotagent.knowledgebase.document_loaders.file.epub.base.EpubReader",
                "flat_pdf": "dotagent.knowledgebase.document_loaders.file.flat_pdf.base.FlatPdfReader",
                "image": "dotagent.knowledgebase.document_loaders.file.image.base.ImageReader",
                "image_caption": "dotagent.knowledgebase.document_loaders.file.image_blip.base.ImageCaptionReader",
                "image_vision": "dotagent.knowledgebase.document_loaders.file.image_blip2.base.ImageVisionLLMReader",
                "image_tabular_chart": "dotagent.knowledgebase.document_loaders.file.image_deplot.base.ImageTabularChartReader",
                "ipynb": "dotagent.knowledgebase.document_loaders.file.ipynb.base.IPYNBReader",
                "json": "dotagent.knowledgebase.document_loaders.file.json.base.JSONReader",
                "markdown": "dotagent.knowledgebase.document_loaders.file.markdown.base.MarkdownReader",
                "mbox": "dotagent.knowledgebase.document_loaders.file.mbox.base.MboxReader",
                "paged_csv": "dotagent.knowledgebase.document_loaders.file.paged_csv.base.PagedCSVReader",
                "pandas_csv": "dotagent.knowledgebase.document_loaders.file.pandas_csv.base.PandasCSVReader",
                "pandas_excel": "dotagent.knowledgebase.document_loaders.file.pandas_excel.base.PandasExcelReader",
                "pdf": "dotagent.knowledgebase.document_loaders.file.pdf.base.PDFReader",
                "pdf_miner": "dotagent.knowledgebase.document_loaders.file.pdf_miner.base.PDFMinerReader",
                "pptx": "dotagent.knowledgebase.document_loaders.file.pptx.base.PptxReader",
                "pymu_pdf": "dotagent.knowledgebase.document_loaders.file.pymu_pdf.base.PyMuPDFReader",
                "rdf": "dotagent.knowledgebase.document_loaders.file.rdf.base.RDFReader",
                "simple_csv": "dotagent.knowledgebase.document_loaders.file.simple_csv.base.SimpleCSVReader",
                "unstructured": "dotagent.knowledgebase.document_loaders.file.unstructured.base.UnstructuredReader",
                "firebase_realtimedb": "dotagent.knowledgebase.document_loaders.firebase_realtimedb.base.FirebaseRealtimeDatabaseReader",
                "firestore": "dotagent.knowledgebase.document_loaders.firestore.base.FirestoreReader",
                "github_repo_issues": "dotagent.knowledgebase.document_loaders.github_repo_issues.base.GitHubRepositoryIssuesReader",
                "gmail": "dotagent.knowledgebase.document_loaders.gmail.base.GmailReader",
                "google_calendar": "dotagent.knowledgebase.document_loaders.google_calendar.base.GoogleCalendarReader",
                "google_docs": "dotagent.knowledgebase.document_loaders.google_docs.base.GoogleDocsReader",
                "google_keep": "dotagent.knowledgebase.document_loaders.google_keep.base.GoogleKeepReader",
                "google_sheets": "dotagent.knowledgebase.document_loaders.google_sheets.base.GoogleSheetsReader",
                "gpt_repo": "dotagent.knowledgebase.document_loaders.gpt_repo.base.GPTRepoReader",
                "graphdb_cypher": "dotagent.knowledgebase.document_loaders.graphdb_cypher.base.GraphDBCypherReader",
                "graphql": "dotagent.knowledgebase.document_loaders.graphql.base.GraphQLReader",
                "hatena_blog": "dotagent.knowledgebase.document_loaders.hatena_blog.base.HatenaBlogReader",
                "hubspot": "dotagent.knowledgebase.document_loaders.hubspot.base.HubspotReader",
                "huggingface_fs": "dotagent.knowledgebase.document_loaders.huggingface.fs.base.HuggingFaceFSReader",
                "intercom": "dotagent.knowledgebase.document_loaders.intercom.base.IntercomReader",
                "jira": "dotagent.knowledgebase.document_loaders.jira.base.JiraReader",
                # "joplin": "dotagent.knowledgebase.document_loaders.joplin.base.JoplinReader",
                "jsondata": "dotagent.knowledgebase.document_loaders.jsondata.base.JSONDataReader",
                "kaltura_esearch": "dotagent.knowledgebase.document_loaders.kaltura.esearch.base.KalturaESearchReader",
                "kibela": "dotagent.knowledgebase.document_loaders.kibela.base.KibelaReader",
                # "make_com": "dotagent.knowledgebase.document_loaders.make_com.base.MakeWrapper",
                "mangoapps_guides": "dotagent.knowledgebase.document_loaders.mangoapps_guides.base.MangoppsGuidesReader",
                "maps": "dotagent.knowledgebase.document_loaders.maps.base.OpenMap",
                "memos": "dotagent.knowledgebase.document_loaders.memos.base.MemosReader",
                "metal": "dotagent.knowledgebase.document_loaders.metal.base.MetalReader",
                "milvus": "dotagent.knowledgebase.document_loaders.milvus.base.MilvusReader",
                "mondaydotcom": "dotagent.knowledgebase.document_loaders.mondaydotcom.base.MondayReader",
                "mongo": "dotagent.knowledgebase.document_loaders.mongo.base.SimpleMongoReader",
                "notion": "dotagent.knowledgebase.document_loaders.notion.base.NotionPageReader",
                "obsidian": "dotagent.knowledgebase.document_loaders.obsidian.base.ObsidianReader",
                "opendal": "dotagent.knowledgebase.document_loaders.opendal_reader.base.OpendalReader",
                "opendal_azblob": "dotagent.knowledgebase.document_loaders.opendal_reader.azblob.base.OpendalAzblobReader",
                "opendal_gcs": "dotagent.knowledgebase.document_loaders.opendal_reader.gcs.base.OpendalGcsReader",
                "opendal_s3": "dotagent.knowledgebase.document_loaders.opendal_reader.s3.base.OpendalS3Reader",
                "outlook_localcalendar": "dotagent.knowledgebase.document_loaders.outlook_localcalendar.base.OutlookLocalCalendarReader",
                "pubmed": "dotagent.knowledgebase.document_loaders.papers.pubmed.base.PubmedReader",
                "pinecone": "dotagent.knowledgebase.document_loaders.pinecone.base.PineconeReader",
                "qdrant": "dotagent.knowledgebase.document_loaders.qdrant.base.QdrantReader",
                "readwise": "dotagent.knowledgebase.document_loaders.readwise.base.ReadwiseReader",
                "reddit": "dotagent.knowledgebase.document_loaders.reddit.base.RedditReader",
                "slack": "dotagent.knowledgebase.document_loaders.slack.base.SlackReader",
                "snscrape_twitter": "dotagent.knowledgebase.document_loaders.snscrape_twitter.base.SnscrapeTwitterReader",
                "spotify": "dotagent.knowledgebase.document_loaders.spotify.base.SpotifyReader",
                "stackoverflow": "dotagent.knowledgebase.document_loaders.stackoverflow.base.StackoverflowReader",
                "steamship": "dotagent.knowledgebase.document_loaders.steamship.base.SteamshipFileReader",
                "string_iterable": "dotagent.knowledgebase.document_loaders.string_iterable.base.StringIterableReader",
                "trello": "dotagent.knowledgebase.document_loaders.trello.base.TrelloReader",
                "twitter": "dotagent.knowledgebase.document_loaders.twitter.base.TwitterTweetReader",
                "weather": "dotagent.knowledgebase.document_loaders.weather.base.WeatherReader",
                "weaviate": "dotagent.knowledgebase.document_loaders.weaviate.base.WeaviateReader",
                "async_web": "dotagent.knowledgebase.document_loaders.web.async_web.base.AsyncWebPageReader",
                "beautiful_soup_web": "dotagent.knowledgebase.document_loaders.web.beautiful_soup_web.base.BeautifulSoupWebReader",
                "knowledge_base_web": "dotagent.knowledgebase.document_loaders.web.knowledge_base.base.KnowledgeBaseWebReader",
                # "readability_web": "dotagent.knowledgebase.document_loaders.web.readability_web.base.ReadabilityWebPageReader",
                "rss": "dotagent.knowledgebase.document_loaders.web.rss.base.RssReader",
                "simple_web": "dotagent.knowledgebase.document_loaders.web.simple_web.base.SimpleWebPageReader",
                # "sitemap": "dotagent.knowledgebase.document_loaders.web.sitemap.base.SitemapReader",
                "trafilatura_web": "dotagent.knowledgebase.document_loaders.web.trafilatura_web.base.TrafilaturaWebReader",
                "unstructured_web": "dotagent.knowledgebase.document_loaders.web.unstructured_web.base.UnstructuredURLLoader",
                "whatsapp": "dotagent.knowledgebase.document_loaders.whatsapp.base.WhatsappChatLoader",
                "wikipedia": "dotagent.knowledgebase.document_loaders.wikipedia.base.WikipediaReader",
                "wordlift": "dotagent.knowledgebase.document_loaders.wordlift.base.WordLiftLoader",
                "wordpress": "dotagent.knowledgebase.document_loaders.wordpress.base.WordpressReader",
                "youtube_transcript": "dotagent.knowledgebase.document_loaders.youtube_transcript.base.YoutubeTranscriptReader",
                "zendesk": "dotagent.knowledgebase.document_loaders.zendesk.base.ZendeskReader",
                "zulip": "dotagent.knowledgebase.document_loaders.zulip.base.ZulipReader",
}
    reader_class = import_class(mapping[reader_type])
    return reader_class


"""
Local implementation of AgentBox.
This is useful for testing and development.c
In case you don't put an api_key,
this is the default AgentBox.
"""

import asyncio
from queue import Queue
import threading
import json
import os
import subprocess
import sys
import time
from asyncio.subprocess import Process
from pathlib import Path
from typing import List, Optional, Union
from uuid import uuid4
import aiohttp
import requests
from websockets.client import WebSocketClientProtocol
from websockets.client import connect as ws_connect
from websockets.exceptions import ConnectionClosedError
from websockets.sync.client import ClientConnection
from websockets.sync.client import connect as ws_connect_sync
from dotagent.agentbox import BaseBox
from dotagent.schema import AgentBoxFile, AgentBoxOutput, AgentBoxStatus
from dotagent.config import settings


class TinyBox(BaseBox):
    """
    TinyBox is a AgentBox implementation that runs code locally.
    This is useful for testing and development.
    """

    _instance: Optional["TinyBox"] = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls, *args, **kwargs)
        else:
            if settings.SHOW_INFO:
                print(
                    "INFO: Using a TinyBox which is not fully isolated\n"
                    "      and not scalable across multiple users.\n"
                    "      Make sure to use a AGENTBOX_API_KEY in production.\n"
                    "      Set envar SHOW_INFO=False to not see this again.\n"
                )
        return cls._instance

    def __init__(self) -> None:
        super().__init__()
        self.port: int = 8888
        self.kernel_id: Optional[dict] = None
        self.ws: Union[WebSocketClientProtocol, ClientConnection, None] = None
        self.jupyter: Union[Process, subprocess.Popen, None] = None
        self.session: Optional[aiohttp.ClientSession] = None
        self.message_queue = Queue()
        self.message_thread = threading.Thread(target=self._message_handler)
        self.message_thread.start()
        
    def _message_handler(self):
        while True:
            message = self.message_queue.get()
            if message == "STOP":
                break
            self._send_message_to_kernel(message)


    def _send_message_to_kernel(self, message):
        msg_json = json.dumps(message)
        if self.ws:
            try:
                self.ws.send(msg_json)
            except ConnectionClosedError:
                # Handle connection closed error if needed
                pass


    def start(self) -> AgentBoxStatus:
        os.makedirs(".agentbox", exist_ok=True)
        self._check_port()
        if settings.VERBOSE:
            print("Starting kernel...")
            out = None
        else:
            out = subprocess.PIPE
        self._check_installed()
        try:
            python = Path(sys.executable).absolute()
            self.jupyter = subprocess.Popen(
                [
                    python,
                    "-m",
                    "jupyter",
                    "kernelgateway",
                    "--KernelGatewayApp.ip='0.0.0.0'",
                    f"--KernelGatewayApp.port={self.port}",
                ],
                stdout=out,
                stderr=out,
                cwd=".agentbox",
            )
        except FileNotFoundError:
            raise ModuleNotFoundError(
                "Jupyter Kernel Gateway not found, please install it with:\n"
                "`pip install jupyter_kernel_gateway`\n"
                "to use the TinyBox."
            )
        while True:
            try:
                response = requests.get(self.kernel_url, timeout=90)
                if response.status_code == 200:
                    break
            except requests.exceptions.ConnectionError:
                pass
            if settings.VERBOSE:
                print("Waiting for kernel to start...")
            time.sleep(1)

        response = requests.post(
            f"{self.kernel_url}/kernels",
            headers={"Content-Type": "application/json"},
            timeout=90,
        )
        self.kernel_id = response.json()["id"]
        if self.kernel_id is None:
            raise Exception("Could not start kernel")

        self.ws = ws_connect_sync(f"{self.ws_url}/kernels/{self.kernel_id}/channels")

        return AgentBoxStatus(status="started")

    def _check_port(self) -> None:
        try:
            response = requests.get(f"http://localhost:{self.port}", timeout=90)
        except requests.exceptions.ConnectionError:
            pass
        else:
            if response.status_code == 200:
                self.port += 1
                self._check_port()

    def _check_installed(self) -> None:
        """if jupyter-kernel-gateway is installed"""
        import pkg_resources  # type: ignore

        try:
            pkg_resources.get_distribution("jupyter-kernel-gateway")
        except pkg_resources.DistributionNotFound:
            print(
                "Make sure 'jupyter-kernel-gateway' is installed "
                "when using without a AGENTBOX_API_KEY.\n"
                "You can install it with 'pip install jupyter-kernel-gateway'."
            )
            raise

    async def astart(self) -> AgentBoxStatus:
        os.makedirs(".agentbox", exist_ok=True)
        self.session = aiohttp.ClientSession()
        await self._acheck_port()
        if settings.VERBOSE:
            print("Starting kernel...")
            out = None
        else:
            out = asyncio.subprocess.PIPE
        self._check_installed()
        python = Path(sys.executable).absolute()
        try:
            self.jupyter = await asyncio.create_subprocess_exec(
                python,
                "-m",
                "jupyter",
                "kernelgateway",
                "--KernelGatewayApp.ip='0.0.0.0'",
                f"--KernelGatewayApp.port={self.port}",
                stdout=out,
                stderr=out,
                cwd=".agentbox",
            )
        except Exception as e:
            print(e)
            raise ModuleNotFoundError(
                "Jupyter Kernel Gateway not found, please install it with:\n"
                "`pip install jupyter_kernel_gateway`\n"
                "to use the TinyBox."
            )
        while True:
            try:
                response = await self.session.get(self.kernel_url)
                if response.status == 200:
                    break
            except aiohttp.ClientConnectorError:
                pass
            except aiohttp.ServerDisconnectedError:
                pass
            if settings.VERBOSE:
                print("Waiting for kernel to start...")
            await asyncio.sleep(1)

        response = await self.session.post(
            f"{self.kernel_url}/kernels", headers={"Content-Type": "application/json"}
        )
        self.kernel_id = (await response.json())["id"]
        if self.kernel_id is None:
            raise Exception("Could not start kernel")
        self.ws = await ws_connect(f"{self.ws_url}/kernels/{self.kernel_id}/channels")

        return AgentBoxStatus(status="started")

    async def _acheck_port(self) -> None:
        try:
            if self.session is None:
                self.session = aiohttp.ClientSession()
            response = await self.session.get(f"http://localhost:{self.port}")
        except aiohttp.ClientConnectorError:
            pass
        except aiohttp.ServerDisconnectedError:
            pass
        else:
            if response.status == 200:
                self.port += 1
                await self._acheck_port()

    def status(self) -> AgentBoxStatus:
        return AgentBoxStatus(
            status="running"
            if self.kernel_id
            and requests.get(self.kernel_url, timeout=90).status_code == 200
            else "stopped"
        )

    async def astatus(self) -> AgentBoxStatus:
        return AgentBoxStatus(
            status="running"
            if self.kernel_id
            and self.session
            and (await self.session.get(self.kernel_url)).status == 200
            else "stopped"
        )

    def run(
        self,
        code: Optional[str] = None,
        file_path: Optional[os.PathLike] = None,
        retry=3,
    ) -> AgentBoxOutput:
        if not code and not file_path:
            raise ValueError("Code or file_path must be specified!")

        if code and file_path:
            raise ValueError("Can only specify code or the file to read_from!")

        if file_path:
            with open(file_path, "r", encoding="utf-8") as f:
                code = f.read()

        # run code in jupyter kernel
        if retry <= 0:
            raise RuntimeError("Could not connect to kernel")
        if not self.ws:
            self.start()
            if not self.ws:
                raise RuntimeError("Could not connect to kernel")
            # Message to send to kernel
        message = {
            "header": {
                "msg_id": (msg_id := uuid4().hex),
                "msg_type": "execute_request",
            },
            "parent_header": {},
            "metadata": {},
            "content": {
                "code": code,
                "silent": False,
                "store_history": True,
                "user_expressions": {},
                "allow_stdin": False,
                "stop_on_error": True,
            },
            "channel": "shell",
            "buffers": [],
        }

        # Put the message into the queue
        self.message_queue.put(message)

        if settings.VERBOSE:
            print("Running code:\n", code)

        # send code to kernel
        self.ws.send(json.dumps(self.message_queue.get()))

        result = ""
        while True:
            try:
                if isinstance(self.ws, WebSocketClientProtocol):
                    raise RuntimeError("Mixing asyncio and sync code is not supported")
                received_msg = json.loads(self.ws.recv())
            except ConnectionClosedError:
                self.start()
                return self.run(code, file_path, retry - 1)

            if (
                received_msg["header"]["msg_type"] == "stream"
                and received_msg["parent_header"]["msg_id"] == msg_id
            ):
                msg = received_msg["content"]["text"].strip()
                if "Requirement already satisfied:" in msg:
                    continue
                result += msg + "\n"
                if settings.VERBOSE:
                    print("Output:\n", result)

            elif (
                received_msg["header"]["msg_type"] == "execute_result"
                and received_msg["parent_header"]["msg_id"] == msg_id
            ):
                result += received_msg["content"]["data"]["text/plain"].strip() + "\n"
                if settings.VERBOSE:
                    print("Output:\n", result)

            elif received_msg["header"]["msg_type"] == "display_data":
                if "image/png" in received_msg["content"]["data"]:
                    return AgentBoxOutput(
                        type="image/png",
                        content=received_msg["content"]["data"]["image/png"],
                    )
                if "text/plain" in received_msg["content"]["data"]:
                    return AgentBoxOutput(
                        type="text",
                        content=received_msg["content"]["data"]["text/plain"],
                    )
                return AgentBoxOutput(
                    type="error",
                    content="Could not parse output",
                )
            elif (
                received_msg["header"]["msg_type"] == "status"
                and received_msg["parent_header"]["msg_id"] == msg_id
                and received_msg["content"]["execution_state"] == "idle"
            ):
                if len(result) > 500:
                    result = "[...]\n" + result[-500:]
                return AgentBoxOutput(
                    type="text", content=result or "code run successfully (no output)"
                )

            elif (
                received_msg["header"]["msg_type"] == "error"
                and received_msg["parent_header"]["msg_id"] == msg_id
            ):
                error = (
                    f"{received_msg['content']['ename']}: "
                    f"{received_msg['content']['evalue']}"
                )
                if settings.VERBOSE:
                    print("Error:\n", error)
                return AgentBoxOutput(type="error", content=error)

    async def arun(
        self,
        code: str,
        file_path: Optional[os.PathLike] = None,
        retry=3,
    ) -> AgentBoxOutput:
        if file_path:
            raise NotImplementedError(
                "Reading from file is not supported in async mode"
            )

        # run code in jupyter kernel
        if retry <= 0:
            raise RuntimeError("Could not connect to kernel")
        if not self.ws:
            await self.astart()
            if not self.ws:
                raise RuntimeError("Could not connect to kernel")
        # Message to send to kernel
        message = {
            "header": {
                "msg_id": (msg_id := uuid4().hex),
                "msg_type": "execute_request",
            },
            "parent_header": {},
            "metadata": {},
            "content": {
                "code": code,
                "silent": False,
                "store_history": True,
                "user_expressions": {},
                "allow_stdin": False,
                "stop_on_error": True,
            },
            "channel": "shell",
            "buffers": [],
        }

        # Put the message into the queue
        self.message_queue.put(message)

        if settings.VERBOSE:
            print("Running code:\n", code)


        if not isinstance(self.ws, WebSocketClientProtocol):
            raise RuntimeError("Mixing asyncio and sync code is not supported")
        
        await self.ws.send(json.dumps(self.message_queue.get()))
        result = ""
        while True:
            try:
                received_msg = json.loads(await self.ws.recv())
            except ConnectionClosedError:
                await self.astart()
                return await self.arun(code, file_path, retry - 1)

            if (
                received_msg["header"]["msg_type"] == "stream"
                and received_msg["parent_header"]["msg_id"] == msg_id
            ):
                msg = received_msg["content"]["text"].strip()
                if "Requirement already satisfied:" in msg:
                    continue
                result += msg + "\n"
                if settings.VERBOSE:
                    print("Output:\n", result)

            elif (
                received_msg["header"]["msg_type"] == "execute_result"
                and received_msg["parent_header"]["msg_id"] == msg_id
            ):
                result += received_msg["content"]["data"]["text/plain"].strip() + "\n"
                if settings.VERBOSE:
                    print("Output:\n", result)

            elif received_msg["header"]["msg_type"] == "display_data":
                if "image/png" in received_msg["content"]["data"]:
                    return AgentBoxOutput(
                        type="image/png",
                        content=received_msg["content"]["data"]["image/png"],
                    )
                if "text/plain" in received_msg["content"]["data"]:
                    return AgentBoxOutput(
                        type="text",
                        content=received_msg["content"]["data"]["text/plain"],
                    )
            elif (
                received_msg["header"]["msg_type"] == "status"
                and received_msg["parent_header"]["msg_id"] == msg_id
                and received_msg["content"]["execution_state"] == "idle"
            ):
                if len(result) > 500:
                    result = "[...]\n" + result[-500:]
                return AgentBoxOutput(
                    type="text", content=result or "code run successfully (no output)"
                )

            elif (
                received_msg["header"]["msg_type"] == "error"
                and received_msg["parent_header"]["msg_id"] == msg_id
            ):
                error = (
                    f"{received_msg['content']['ename']}: "
                    f"{received_msg['content']['evalue']}"
                )
                if settings.VERBOSE:
                    print("Error:\n", error)
                return AgentBoxOutput(type="error", content=error)

    def upload(self, file_name: str, content: bytes) -> AgentBoxStatus:
        os.makedirs(".agentbox", exist_ok=True)
        with open(os.path.join(".agentbox", file_name), "wb") as f:
            f.write(content)

        return AgentBoxStatus(status=f"{file_name} uploaded successfully")

    async def aupload(self, file_name: str, content: bytes) -> AgentBoxStatus:
        return await asyncio.to_thread(self.upload, file_name, content)

    def download(self, file_name: str) -> AgentBoxFile:
        with open(os.path.join(".agentbox", file_name), "rb") as f:
            content = f.read()

        return AgentBoxFile(name=file_name, content=content)

    async def adownload(self, file_name: str) -> AgentBoxFile:
        return await asyncio.to_thread(self.download, file_name)

    def install(self, package_name: str) -> AgentBoxStatus:
        self.run(f"!pip install -q {package_name}")
        
        return AgentBoxStatus(status=f"{package_name} installed successfully")

    async def ainstall(self, package_name: str) -> AgentBoxStatus:
        await self.arun(f"!pip install -q {package_name}")
        
        return AgentBoxStatus(status=f"{package_name} installed successfully")

    def list_files(self) -> List[AgentBoxFile]:
        return [
            AgentBoxFile(name=file_name, content=None)
            for file_name in os.listdir(".agentbox")
        ]

    async def alist_files(self) -> List[AgentBoxFile]:
        return await asyncio.to_thread(self.list_files)

    def restart(self) -> AgentBoxStatus:
        if self.jupyter is not None:
            self.stop()
        else:
            self.start()
        return AgentBoxStatus(status="restarted")

    async def arestart(self) -> AgentBoxStatus:
        if self.jupyter is not None:
            await self.astop()
        else:
            await self.astart()
        return AgentBoxStatus(status="restarted")

    def stop(self) -> AgentBoxStatus:
        self.message_queue.put("STOP")
        self.message_thread.join()
        if self.ws is not None:
            try:
                self.ws.close()
            except ConnectionClosedError:
                pass
            self.ws = None

        if self.jupyter is not None:
            self.jupyter.terminate()
            self.jupyter.wait()
            self.jupyter = None
            time.sleep(2)

        return AgentBoxStatus(status="stopped")

    async def astop(self) -> AgentBoxStatus:
        self.message_queue.put("STOP")
        self.message_thread.join()  
        if self.ws is not None:
            try:
                if not isinstance(self.ws, WebSocketClientProtocol):
                    raise RuntimeError("Mixing asyncio and sync code is not supported")
                await self.ws.close()
            except ConnectionClosedError:
                pass
            self.ws = None

        if self.jupyter is not None:
            self.jupyter.terminate()
            self.jupyter = None
            await asyncio.sleep(2)

        if self.session is not None:
            await self.session.close()
            self.session = None

        return AgentBoxStatus(status="stopped")

    @property
    def kernel_url(self) -> str:
        """Return the url of the kernel."""
        return f"http://localhost:{self.port}/api"

    @property
    def ws_url(self) -> str:
        """Return the url of the websocket."""
        return f"ws://localhost:{self.port}/api"



from typing import List

from dotagent.knowledgebase.document_loaders.basereader import BaseReader
from dotagent.schema import DocumentNode


def safe_value_dict(dict_obj):
    for key, value in dict_obj.items():
        if isinstance(value, (str, int, float)):
            dict_obj[key] = value
        elif isinstance(value, list):
            # Convert lists to strings
            dict_obj[key] = ", ".join(map(str, value))
        elif value is None:
            # Replace None with a default string
            dict_obj[key] = ""
        else:
            # Convert other types to strings
            dict_obj[key] = str(value)
    return dict_obj


class JiraReader(BaseReader):
    """Jira reader. Reads data from Jira issues from passed query.

    Args:
        email (str): Jira email.
        api_token (str): Jira API token.
        server_url (str): Jira server url.
    """

    def __init__(self, email: str, api_token: str, server_url: str) -> None:

        from jira import JIRA

        self.jira = JIRA(basic_auth=(email, api_token), server=f"https://{server_url}")

    def load_data(self, query: str) -> List[DocumentNode]:
        relevant_issues = self.jira.search_issues(query)

        issues = []

        for issue in relevant_issues:
            # Iterates through only issues and not epics
            if "parent" in (issue.raw["fields"]):
                assignee = ""
                reporter = ""
                epic_key = ""
                epic_summary = ""
                epic_descripton = ""

                if issue.fields.assignee:
                    assignee = issue.fields.assignee.displayName

                if issue.fields.reporter:
                    reporter = issue.fields.reporter.displayName

                if issue.raw["fields"]["parent"]["key"]:
                    epic_key = issue.raw["fields"]["parent"]["key"]

                if issue.raw["fields"]["parent"]["fields"]["summary"]:
                    epic_summary = issue.raw["fields"]["parent"]["fields"]["summary"]

                if issue.raw["fields"]["parent"]["fields"]["status"]["description"]:
                    epic_descripton = issue.raw["fields"]["parent"]["fields"]["status"][
                        "description"
                    ]

                issues.append(
                    DocumentNode(
                        text=f"{issue.fields.summary} \n {issue.fields.description}",
                        extra_info=safe_value_dict(
                            {
                                "id": issue.id,
                                "title": issue.fields.summary,
                                "url": issue.permalink(),
                                "query": query,
                                "created_at": issue.fields.created,
                                "updated_at": issue.fields.updated,
                                "labels": issue.fields.labels,
                                "status": issue.fields.status.name,
                                "assignee": assignee,
                                "reporter": reporter,
                                "project": issue.fields.project.name,
                                "issue_type": issue.fields.issuetype.name,
                                "priority": issue.fields.priority.name,
                                "epic_key": epic_key,
                                "epic_summary": epic_summary,
                                "epic_description": epic_descripton,
                            }
                        ),
                    )
                )

        return issues



from typing import List

import numpy as np
from pydantic import BaseModel

from dotagent.vectorstores.embeddings.base import Embeddings


class FakeEmbeddings(Embeddings, BaseModel):
    size: int

    def _get_embedding(self) -> List[float]:
        return list(np.random.normal(size=self.size))

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        return [self._get_embedding() for _ in texts]

    def embed_query(self, text: str) -> List[float]:
        return self._get_embedding()


"""Notion tool spec."""

from dotagent.tools.toolkits.base import BaseToolkit
from dotagent.tools.toolkits.notion_toolkit.notion.base import LoadData, SearchData
from dotagent.tools.basetool import BaseTool
from typing import Optional, List

SEARCH_URL = "https://api.notion.com/v1/search"


class NotionToolkit(BaseToolkit):
    """Notion tool spec.

    Currently a simple wrapper around the data loader.
    TODO: add more methods to the Notion spec.

    """
    integration_token: Optional[str] = None

    def get_tools(self) -> List[BaseTool]:
        """Get the tools in the toolkit."""
        return [
            LoadData(integration_token = self.integration_token),
            SearchData(integration_token = self.integration_token)
        ]


from typing import Any, List, Optional

from dotagent.schema import BaseRetriever, Document


class MetalRetriever(BaseRetriever):
    def __init__(self, client: Any, params: Optional[dict] = None):
        try: 
            from metal_sdk.metal import Metal
        except ImportError:
            raise ImportError("Could not import Metal package. Please install it with 'pip install metal'")

        if not isinstance(client, Metal):
            raise ValueError(
                "Got unexpected client, should be of type metal_sdk.metal.Metal. "
                f"Instead, got {type(client)}"
            )
        self.client: Metal = client
        self.params = params or {}

    def get_relevant_documents(self, query: str) -> List[Document]:
        results = self.client.search({"text": query}, **self.params)
        final_results = []
        for r in results["data"]:
            metadata = {k: v for k, v in r.items() if k != "text"}
            final_results.append(Document(page_content=r["text"], metadata=metadata))
        return final_results

    async def aget_relevant_documents(self, query: str) -> List[Document]:
        raise NotImplementedError("Metal retriever does not support async")


from dotagent.utils.utils import (
    get_from_dict_or_env,
    get_from_env,
    stringify_dict,
    stringify_value,
    comma_list
)

__all__ = ["get_from_dict_or_env", "get_from_env", "stringify_dict","stringify_value","comma_list"]



import asyncio
import builtins
from .._utils import ContentCapture

async def each(list, hidden=False, parallel=False, item_name="this", start_index=0, _parser_context=None):
    ''' Iterate over a list and execute a block for each item.

    Parameters
    ----------
    list : iterable
        The list to iterate over. Inside the block each element will be available as `this`.
    hidden : bool
        Whether to include the generated item blocks in future LLM context.
    parallel : bool
        If this is True then we generate all the items in the list in parallel. Note that this
        is only compatible with hidden=True. When parallel=True you can no longer raise a
        StopIteration exception to stop the loop at a specific step (since the steps can be run
        in parallel in any order).
    item_name : str
        The name of the variable to use for the current item in the list.
    '''
    block_content = _parser_context['block_content']
    parser = _parser_context['parser']
    variable_stack = _parser_context['variable_stack']

    assert not parallel or hidden is True, "parallel=True is only compatible with hidden=True (since if hidden=False earlier items are contex for later items)"

    if isinstance(list, dict):
        keys = builtins.list(list.keys())
        list = list.values()
    else:
        keys = None

    # make sure the list is iterable
    try:
        iter(list)
    except TypeError:
        raise TypeError("The #each command cannot iterate over a non-iterable value: " + str(list))

    out = []

    # we parse in parallel if the loop contents are hidden
    if parallel:
        
         # set up the coroutines to call
        coroutines = []
        for i, item in enumerate(list):
            if i < start_index: # skip items before the start index
                continue
            context = {
                "@index": i,
                "@first": i == 0,
                "@last": i == len(list) - 1,
                item_name: item,
                "@raw_prefix": variable_stack["@raw_prefix"], # create a local copy of the prefix since we are hidden
                "@no_display": True
            }
            if keys is not None:
                context["@key"] = keys[i]
            variable_stack.push(context)
            coroutines.append(parser.visit(
                block_content,
                variable_stack.copy(),
                next_node=_parser_context["next_node"],
                next_next_node=_parser_context["next_next_node"],
                prev_node=_parser_context["prev_node"]
            ))
            variable_stack.pop()

        await asyncio.gather(*coroutines)
        
            # for item_out in item_outs:

            #     # parser._trim_prefix(item_out)
            #     out.append(item_out)

            #     # check if the block has thrown a stop iteration signal
            #     if parser.caught_stop_iteration:
            #         parser.caught_stop_iteration = False
            #         break
    else:
        for i, item in enumerate(list):
            context = {
                "@index": i,
                "@first": i == 0,
                "@last": i == len(list) - 1,
                item_name: item
            }
            if keys is not None:
                context["@key"] = keys[i]
            variable_stack.push(context)
            with ContentCapture(variable_stack, hidden) as new_content:
                new_content += await parser.visit(
                    block_content,
                    variable_stack,
                    next_node=_parser_context["next_node"],
                    next_next_node=_parser_context["next_next_node"],
                    prev_node=_parser_context["prev_node"]
                )
                out.append(str(new_content))
            variable_stack.pop()

            # if we stopped executing then we need to dump our node text back out but with the start_index incremented to account for what we've already done
            if not parser.executing:
                updated_text = re.sub(r"^({{~?#each.*?)(~?}})", r"\1 start_index="+str(i+1)+r"\2", _parser_context["parser_node"].text)
                variable_stack["@raw_prefix"] += updated_text
                break

            # check if the block has thrown a stop iteration signal
            if parser.caught_stop_iteration:
                parser.caught_stop_iteration = False
                break
        
            
    
    # if not hidden:
        #return "{{!--GMARKER_each$$--}}" + "{{!--GMARKER_each$$--}}".join(out) + "{{!--GMARKER_each$$--}}" + suffix
    # if hidden:
    #     id = uuid.uuid4().hex
    #     l = len(out)
    #     out_str = "{{!--" + f"GMARKER_each_noecho_start_{not hidden}_{l}${id}$" + "--}}"
    #     for i, value in enumerate(out):
    #         if i > 0:
    #             out_str += "{{!--" + f"GMARKER_each_noecho_{not hidden}_{i}${id}$" + "--}}"
    #         out_str += value
    #     return out_str + "{{!--" + f"GMARKER_each_noecho_end${id}$" + "--}}"

each.is_block = True


import os
import requests
import base64

from ._base import BaseImageLlm


class StableDiffusion(BaseImageLlm):
    def __init__(self, api_key, image_model=None, number_of_results=1, client_id=None, client_version=None):
        """
        Args:
            api_key (str): The Stability API key.
            image_model (str): The image model.
            number_of_results (int): The number of results.
            client_id (str): Client ID.
            client_version (str): Client version.
        """
        self.api_key = api_key
        self.image_model = image_model or 'stable-diffusion-xl-beta-v2-2-2'
        self.number_of_results = number_of_results
        self.api_host = os.getenv('API_HOST', 'https://api.stability.ai')
        self.url = f"{self.api_host}/v1/generation/{self.image_model}/text-to-image"
        self.client_id = client_id
        self.client_version = client_version

    def get_image_model(self):
        """
        Returns:
            str: The image model.
        """
        return self.image_model

    def generate_image(self, prompt: str, size: int = 512, style_preset='enhance', cfg_scale=7, steps=50, seed=0):
        """
        Call the Stability image API.

        Args:
            prompt (str): The prompt.
            size (int): The size.
            style_preset (str): The style preset.
            cfg_scale (int): The config scale.
            steps (int): The number of diffusion steps.
            seed (int): The seed for random noise.

        Returns:
            dict: The response.
        """
        body = {
            "width": size,
            "height": size,
            "steps": steps,
            "seed": seed,
            "cfg_scale": cfg_scale,
            "samples": self.number_of_results,
            "style_preset": style_preset,
            "text_prompts": [
                {
                "text": prompt,
                "weight": 1
                }
            ],
        }
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
        }

        # Add client ID and version headers if provided
        if self.client_id is not None:
            headers["Stability-Client-ID"] = self.client_id
        if self.client_version is not None:
            headers["Stability-Client-Version"] = self.client_version

        response = requests.post(
            self.url,
            headers=headers,
            json=body,
        )

        if response.status_code != 200:
            raise Exception("Non-200 response: " + str(response.text))

        data = response.json()

        for i, image in enumerate(data["artifacts"]):
            with open(f"./out/txt2img_{image['seed']}.png", "wb") as f:
                f.write(base64.b64decode(image["base64"]))

        return data



from .._transformers import Transformers
import transformers

class LLaMA2(Transformers):
    """ 
    A HuggingFace transformers version of the LLaMA2 language model with compiler support.
    """

    llm_name: str = "llama2"

    def _model_and_tokenizer(self, model, tokenizer, **kwargs):
        # load the LLaMA2 specific tokenizer and model
        if isinstance(model, str):
            if tokenizer is None:
                tokenizer = transformers.LlamaTokenizer.from_pretrained(
                    model, 
                    add_special_tokens={"pad_token":"<pad>"}, 
                    **kwargs
                )
                tokenizer.model_max_length = transformers.LlamaConfig().max_position_embeddings
            model = transformers.LlamaForCausalLM.from_pretrained(model, **kwargs)
            model.config.pad_token_id = tokenizer.pad_token_id
            
        return super()._model_and_tokenizer(model, tokenizer, **kwargs)
